#ifndef _COMMONS_H_
#define _COMMONS_H_

#include <mex.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <fstream>
#include <cassert>
#include <vector>
#include "matrixOperations.h"
#include "kernels.h"
#include <math.h>

enum kernelType {EPANECHNIKOV=0,GAUSSIAN=1,TUKEY=2};

// perform mean shift iterations
static double* meanShift_cpp(double* X, double* invH, double* mode, int k, double* scale, int npts, kernelType ktype, bool isHeteroscedastic)
{
	double *num, *den, *mshift;
	double *invHi, *Xi;
	double *tempVec1, *tempVec2, *tempVec3, *tempVal4;

	num = (double*)mxCalloc(k,sizeof(double));
	den = (double*)mxCalloc(k*k,sizeof(double));
	invHi = (double*)mxCalloc(k*k,sizeof(double));
	Xi = (double*)mxMalloc(k*sizeof(double));

	tempVec1 = (double*)mxMalloc(k*sizeof(double));
	tempVec2 = (double*)mxMalloc(k*sizeof(double));
	tempVec3 = (double*)mxMalloc(k*sizeof(double));
	tempVal4 = (double*)mxMalloc(1*sizeof(double));

	mshift = (double*)mxMalloc(k*sizeof(double));

	for(int npts_i = 0; npts_i < npts; npts_i++)
	{
		//copy Xi and invHi
		memcpy(Xi, X + npts_i*k, k*sizeof(double));
		for(int j = 0; j < k; j++)
			tempVec2[j] = (mode[j] - Xi[j])/(scale[j]);

		double detHi = 1;
		if(isHeteroscedastic)
		{
			memcpy(invHi, invH + npts_i*k*k, k*k*sizeof(double));
			//compute determinant of Hi
			detHi = 1.0/matrixDeterminant(invHi,k);
			matrixMultiply(invHi,Xi,tempVec1,k,k,1);
			matrixMultiply(tempVec2,invHi,tempVec3,1,k,k);
			matrixMultiply(tempVec3,tempVec2,tempVal4,1,k,1);
		}
		else
		{
			memcpy(tempVec1,Xi,sizeof(double)*k);
			matrixMultiply(tempVec2,tempVec2,tempVal4,1,k,1);
			for(int k_i = 0; k_i < k; k_i++)
				invHi[k_i*(k+1)] = 1.0;
		}

		switch (ktype)
		{
		case EPANECHNIKOV:
			 {
				 for(int k_i = 0; k_i < k; k_i++)
					 num[k_i] += tempVec1[k_i]*gEpanechnikov(*tempVal4)/sqrt(detHi);
				 for(int k_i = 0; k_i < k*k; k_i++)
					 den[k_i] += invHi[k_i]*gEpanechnikov(*tempVal4)/sqrt(detHi);
				 break;
			 }
		case GAUSSIAN:
			 {
				 for(int k_i = 0; k_i < k; k_i++)
					 num[k_i] += tempVec1[k_i]*gGaussian(*tempVal4)/sqrt(detHi);
				 for(int k_i = 0; k_i < k*k; k_i++)
					 den[k_i] += invHi[k_i]*gGaussian(*tempVal4)/sqrt(detHi);
				 break;
			 }
		case TUKEY:
			 {
				 for(int k_i = 0; k_i < k; k_i++)
					 num[k_i] += tempVec1[k_i]*gTukey(*tempVal4)/sqrt(detHi);
				 for(int k_i = 0; k_i < k*k; k_i++)
					 den[k_i] += invHi[k_i]*gTukey(*tempVal4)/sqrt(detHi);
				 break;
			 }
		}
	}

	double detDen = matrixDeterminant(den,k);

	//compute mean shift vector
	if(abs(detDen) >  1.0e-15)
	{
		double *invDen = (double*)mxMalloc(k*k*sizeof(double));
		matrixInverse(den,invDen,k,k);
		matrixMultiply(invDen,num,mshift,k,k,1);
		mxFree(invDen);
		for(int k_i = 0; k_i < k; k_i++)
			mshift[k_i] -= mode[k_i];
	}
	else
	{
		for(int k_i = 0; k_i < k; k_i++)
			mshift[k_i] = 0;
	}

	mxFree(num);
	mxFree(den);
	mxFree(Xi);
	mxFree(invHi);
	mxFree(tempVec1);
	mxFree(tempVec2);
	mxFree(tempVec3);
	mxFree(tempVal4);

	return mshift;
}

// compute Mahalanobis Distance
static double computeMahalanobisDistance(double* diff, double* invH_i, int k, bool isHeteroscedastic)
{
	double* tempVec = (double*)mxMalloc(k*sizeof(double));
	int tempVec_dim[2] = {k,1};
	int diff_dim[2] = {k,1};
	double dist_i;

	if(isHeteroscedastic)
	{
		matrixMultiply(diff,invH_i,tempVec,1,k,k);
		innerProduct(tempVec,tempVec_dim,diff,diff_dim,&dist_i);
	}
	else
		innerProduct(diff,diff_dim,diff,diff_dim,&dist_i);

	mxFree(tempVec);

	return dist_i;
}

//Compute best carrier per point
static void computeBestCarriers(double* proj_all, double* proj_best, double* invH, double* alpha, double* dists, int* ind_constraints, int npts, int nCarriers, int k, bool isHeteroscedastic)
{
	double* diff = (double*)mxMalloc(k*sizeof(double));

	for(int pt_i = 0; pt_i < npts*nCarriers; pt_i += nCarriers)
	{
		double dist_mahal = 0.0;
		double dist_mahal_best = -100.0;

		for(int const_i = 0; const_i < nCarriers; const_i++)
		{
			for(int dim_i = 0; dim_i < k; dim_i++)
				diff[dim_i] = proj_all[(pt_i + const_i)*k + dim_i] - alpha[dim_i];

			dist_mahal = computeMahalanobisDistance(diff,&invH[k*k*(pt_i + const_i)],k,isHeteroscedastic);
			if(dist_mahal < 0.0)
				int debug_p = 100;

			if(dist_mahal > dist_mahal_best)
			{
				dist_mahal_best = dist_mahal;
				ind_constraints[pt_i/nCarriers] = const_i;
			}
		}
		dists[pt_i/nCarriers] = dist_mahal_best;

		for(int dim_i = 0; dim_i < k; dim_i++)
			proj_best[(int)(pt_i/nCarriers)*k + dim_i] = proj_all[(pt_i + ind_constraints[pt_i/nCarriers])*k + dim_i];
	}
	mxFree(diff);
}

// routines for quick sort
static int compare(const void *X, const void *Y)
{
	double x = *((double *)X);
	double y = *((double *)Y);

	if (x >= y)
		return 1;
	else
		return -1;
}

static double partition(double a[], int ind[], int left, int right)
{
	double pivot = a[left];
	while (true)
	{
		while (a[left] < pivot) left++;
		while (a[right] > pivot) right--;
		if (left < right)
		{
			swap(a[left], a[right]);
			swap(ind[left], ind[right]);
		}
		else
		{
			return a[right];
		}
	}
}

static void quicksort(double a[], int ind[], int left, int right)
{
	if (left < right)
	{
		double pivot = partition(a, ind, left, right);
		quicksort(a, ind, left, pivot-1);
		quicksort(a, ind, pivot+1, right);
	}
}

static void quickSort(double arr[], int ind[], int left, int right) {

	int i = left, j = right;
	double tmp;
	double pivot = arr[(left + right) / 2];

	/* partition */
	while (i <= j) {
		while (arr[i] < pivot)
			i++;
		while (arr[j] > pivot)
			j--;
		if (i <= j) {
			tmp = arr[i];
			arr[i] = arr[j];
			arr[j] = tmp;
			int temp = ind[i];
			ind[i] = ind[j];
			ind[j] = temp;
			i++;
			j--;
		}
	};

	/* recursion */
	if (left < j)
		quickSort(arr, ind, left, j);
	if (i < right)
		quickSort(arr, ind, i, right);
}

#endif