#include "conjugateGradient.h"

extern "C" int dgesvd_(char*, char*, int*, int*, double*, int*, double*,
					   double*, int*, double*, int*, double*, int*, int*);

conjugateGradient::conjugateGradient()
{
	hasData = 0;
}

// Destructor
conjugateGradient::~conjugateGradient()
{
	clearData();
}

void conjugateGradient::initializeData(double* data, int n_, int N_, int nCarriers_, int npts_, int k_, double* scale_, double* invH_, kernelType kType_, bool isHeteroscedastic_, bool isHomogenous_)
{
	clearData();

	hasData = 1;

	kType = kType_;
		
	n = n_;
	k = k_;
	p = (n+1)*k;
	npts = npts_;
	N = N_;
	nCarriers = nCarriers_;

	theta_dim[0] = n; theta_dim[1] = k;
	Y_dim[0] = n; Y_dim[1] = npts*nCarriers;

	dists = new double[npts];
	ind_constraints = new int[npts];

	// Copy data...
	Y = data;

	invH = invH_;
	invH_best = new double[k*k*npts];

	isHeteroscedastic = isHeteroscedastic_;
	isHomogeneous = isHomogenous_;

	// bandwidth
	scale = scale_;

	proj_all	 = new double[npts*nCarriers*k];
	proj_best	 = new double[npts * k];

	// LAPACK and matrix multiplication stuff...
	m_nLwork = max(5 * n, k * k);
	m_dbArrWork  = new double[m_nLwork];
	m_dbArrWork0 = new double[n * k];
	m_dbArrWork1 = new double[n * k];

	// used in Grassmann cg
	m_ThetaAlpha= new double[p];
	m_dbArrT	= new double[p]; 
	m_dbArrW	= new double[p];

	m_G			= new double[p];
	m_dbArrH	= new double[p];
	m_dbArrX	= new double[p];

	m_U			= new double[p];
	m_V			= new double[k * k];
	m_S			= new double[k];
	m_Cos		= new double[k];
	m_Sin		= new double[k];

	m_Vec1 = new double[k];
	m_Vec2 = new double[k];
	m_Vec3 = new double[1];
}

void conjugateGradient::clearData()
{
	if(!hasData)
		return;

	hasData = 0;

	delete [] proj_all;
	delete [] proj_best;

	delete [] dists;
	delete [] ind_constraints;
	delete [] invH_best;

	delete [] m_dbArrWork;	
	delete [] m_dbArrWork0;	
	delete [] m_dbArrWork1;

	delete [] m_ThetaAlpha;		
	delete [] m_dbArrT;		
	delete [] m_dbArrW;
	delete [] m_G;			
	delete [] m_dbArrH;		
	delete [] m_dbArrX;
	delete [] m_U;			
	delete [] m_V;			
	delete [] m_S;
	delete [] m_Cos;		
	delete [] m_Sin;

	delete [] m_Vec1;
	delete [] m_Vec2;
	delete [] m_Vec3;
}

void conjugateGradient::gmbrak(double& ax, double& bx, double& cx, double& fa, double& fb, double& fc)
{
	double shftstorage, r, q, u, ulim, fu;

	fa = gmComputeFuncAlongLine(ax);
	fb = gmComputeFuncAlongLine(bx);

	// Switch a and b if necessary so we can always go downhill from a to b
	if(fb > fa){
		SHFT(shftstorage, ax, bx, shftstorage);
		SHFT(shftstorage, fa, fb, shftstorage);
	}

	// Initial guess for c
	cx = bx + GOLD * (bx - ax);
	fc = gmComputeFuncAlongLine(cx);

	while(fb > fc){ // Keep iterating till a minimum is bracketed

		r = (bx - ax)*(fb - fc);
		q = (bx - cx)*(fb - fa);
		u = bx - ((bx - cx)*q - (bx - ax)*r) / (2 * SIGN(FMAX(fabs(q-r),TINY),q-r));
		ulim = bx + GLIMIT * (cx - bx); // limit for c

		if((bx - u) * (u - cx) > 0.0){ // parabolic u lies between b and c
			fu = gmComputeFuncAlongLine(u);
			if(fu < fc){ // got a minimum between b and c
				ax = bx;
				bx = u;
				fa = fb;
				fb = fu;
				return;
			}
			else if(fu > fb){ // got a minimum between a and u
				cx = u;
				fc = fu;
				return;
			}
			u = cx + GOLD * (cx - bx); // parabolic fit did not work out...use default magnification
			fu = gmComputeFuncAlongLine(u);
		}
		else if((cx - u) * (u - ulim) > 0.0){ // parabolic fit lies between c and its allowed limit
			fu = gmComputeFuncAlongLine(u);
			if(fu < fc){
				SHFT(bx, cx, u, cx + GOLD * (cx - bx));
				SHFT(fb, fc, fu, gmComputeFuncAlongLine(u));
			}
		}
		else if((u - ulim) * (ulim - cx) > 0.0){ // limit parabolic u to itsmaximum allowed value
			u = ulim;
			fu = gmComputeFuncAlongLine(u);
		}
		else{ // reject parabolic u and use default magnification
			u = cx + GOLD * (cx - bx);
			fu = gmComputeFuncAlongLine(u);
		}

		// eliminate oldest point
		SHFT(ax, bx, cx, u);
		SHFT(fa, fb, fc, fu);
	}
}

double conjugateGradient::gmbrent(double ax, double bx, double cx, double tol, double& xmin)
{
	int iter;
	double a, b, d, etemp, fu, fv, fw, fx, p, q, r, tol1, tol2, u, v, w, x, xm, e = 0.0;

	a = (ax < cx ? ax:cx);
	b = (ax > cx ? ax:cx);

	x = w = v = bx;
	fw = fv = fx = gmComputeFuncAlongLine(x);

	for(iter = 0; iter < MAXITER; iter++){

		xm = 0.5*(a + b);
		tol2 = 2.0*(tol1 = tol*fabs(x) + TINY);
		if(fabs(x - xm) <= (tol2 - 0.5*(b - a))){
			xmin = x;
			return fx;
		}
		if(fabs(e) > tol1){		//Construct a trial parabolic fit
			r = (x - w)*(fx - fv);
			q = (x - v)*(fx - fw);
			p = (x - v)*q - (x - w)*r;
			q = 2.0*(q - r);
			if(q > 0.0)
				p = -p;
			q = fabs(q);
			etemp = e;
			e = d;

			if(fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a - x) || p >= q*(b - x))
				d = CGOLD * (e = (x >= xm ? a - x : b - x));
			else
			{
				d = p / q;
				u = x + d;
				if(u - a < tol2 || b - u < tol2)
					d = SIGN(tol1,xm-x);//fabs(tol1) * (xm > x ? 1 : -1);
			}
		}
		else
			d = CGOLD * (e=(x >= xm ? a - x : b - x));

		u = (fabs(d) >= tol1 ? x + d : x + SIGN(tol1,d));
		fu = gmComputeFuncAlongLine(u);
		if(fu <= fx){
			if(u >= x)
				a = x;
			else
				b = x;
			SHFT(v, w, x, u);
			SHFT(fv, fw, fx, fu);
		}
		else{
			if(u < x)
				a = u;
			else
				b = u;
			if(fu <= fw || w == x){
				v = w;
				w = u;
				fv = fw;
				fw = fu;
			}
			else if(fu <= fv || v == x || v == w){
				v = u;
				fv = fu;
			}
		}
	}
	xmin = x;
	return fx;
}

// this function is currently not being used
double conjugateGradient::gmDirMin(double& ax, double& bx, double& cx)
{
	double ux, fu;
	double fa,fb,fc;

	fa = gmComputeFuncAlongLine(ax); 
	fc = gmComputeFuncAlongLine(cx);
	fb = gmComputeFuncAlongLine(bx);

	ux = ax + RGOLD*(cx-ax);
	fu = gmComputeFuncAlongLine(ux);

	int counter = 0;
	while(abs(cx - ax) > TINY && counter < 100)
	{
		counter++;
		if(fb < fu)
		{
			cx = ux;
			ux = bx;
			bx = RGOLD*ax + CGOLD*ux;
		}
		else if(fb > fu)
		{
			ax = bx;
			bx = ux;
			ux = RGOLD*cx + CGOLD*bx;
		}
		fa = gmComputeFuncAlongLine(ax);
		fc = gmComputeFuncAlongLine(cx);
		fb = gmComputeFuncAlongLine(bx);
		fu = gmComputeFuncAlongLine(ux);
//		mexPrintf("\n\tax = %f, bx = %f, cx = %f,\n", ax, bx, cx);
//		mexPrintf("\tfa = %f, fb = %f, fc = %f,\n", fa, fc, fb);
	}
	return(fb);
}


void conjugateGradient::parallelTransport(double *y, double *u, double *v, double *s, double *cs, double *sn, double *tg, double *th)
{
	int i, j, kk, idx;

	// m_dbArrWork0 = y * v * sn
	idx = 0;
	for(i = 0; i < k; i++){
		for(j = 0; j < n; j++){
			for(kk = 0, m_dbArrWork0[idx] = 0; kk < k; kk++)
				m_dbArrWork0[idx] += y[kk * n + j] * v[kk * k + i] * sn[i];
			idx++;
		}
	}

	// m_dbArrWork1 = u * cs
	idx = kk = 0;
	for(j = 0; j < k; j++)
		for(i = 0; i < n; i++)
			m_dbArrWork1[idx++] = u[kk++] * cs[j];
	
	// m_dbArrWork = u^t * g
	idx = 0;
	for(i = 0; i < k; i++){
		for(j = 0; j < k; j++){
			for(kk = 0, m_dbArrWork[idx] = 0; kk < n; kk++)
				m_dbArrWork[idx] += u[j * n + kk] * tg[i * n + kk];
			idx++;
		}
	}

	// compute th and tg
	idx = 0;
	for(i = 0; i < k; i++){
		for(j = 0; j < n; j++){
			for(kk = 0, th[idx] = 0; kk < k; kk++){
				th[idx] += (m_dbArrWork1[kk * n + j] - m_dbArrWork0[kk * n + j]) * s[kk] * v[i * k + kk];
				tg[idx] -= (m_dbArrWork0[kk * n + j] + u[kk * n + j] - m_dbArrWork1[kk * n + j]) * m_dbArrWork[i * k + kk];
			}
			idx++;
		}
	}
}

void conjugateGradient::geodesicMove(double *y, double *u, double *v, double *c, double *s, double *ty)
{
	int i, j, kk;

	// m_dbArrWork1 = v * c * v^t
	for(i = 0; i < k; i++){
		for(j = i; j < k; j++){
			for(kk = 0, m_dbArrWork1[j * k + i] = 0; kk < k; kk++)
				m_dbArrWork1[j * k + i] += v[i * k + kk] * c[kk] * v[j * k + kk];
			m_dbArrWork1[i * k + j] = m_dbArrWork1[j * k + i];
		}
	}

	// ty = y * m_dbArrWork1 + u * s * v^t
	for(i = 0; i < n; i++){
		for(j = 0; j < k; j++){
			for(kk = 0, ty[j * n + i] = 0; kk < k; kk++)
				ty[j * n + i] += y[kk * n + i] * m_dbArrWork1[j * k + kk] + u[kk * n + i] * s[kk] * v[j * k + kk];
		}
	}
}


double conjugateGradient::gmComputeFuncAlongLine(double t)
{
	int i, j, kk;
	double density;

	// compute parallel transported parameters
	for(i = 0; i < k; i++)
	{
		m_Cos[i] = cos(m_S[i] * t);
		m_Sin[i] = sin(m_S[i] * t);
	}

	geodesicMove(m_ThetaAlpha, m_U, m_V, m_Cos, m_Sin, m_dbArrT);
	if(!isHomogeneous)
	{
		for(j = n * k; j < p; j++)
			m_dbArrT[j] = m_ThetaAlpha[j] + m_U[j] * t;
	}
	else
	{
		for(j = n * k; j < p; j++)
			m_dbArrT[j] = m_ThetaAlpha[j];
	}

	// compute all projections
	innerProduct(m_dbArrT,theta_dim,Y,Y_dim,proj_all);

	// compute best projection per carrier
	computeBestCarriers(proj_all, proj_best, invH, &(m_dbArrT[n*k]), dists, ind_constraints, npts, nCarriers, k, isHeteroscedastic);

	if(isHeteroscedastic)
	{
		for(int pt_i = 0; pt_i < npts; pt_i++)
			memcpy(&(invH_best[pt_i*k*k]),&(invH[(pt_i*nCarriers + ind_constraints[pt_i])*k*k]),k*k*sizeof(double));
	}

	// compute density
	for(i = 0, density = 0, pProj = proj_best; i < npts; i++, pProj++)
	{
		for(kk = 0; kk < k; kk++)
			m_Vec1[kk] = (pProj[kk*npts] - m_dbArrT[k*n+kk])/scale[kk];

		if(isHeteroscedastic)
		{
			matrixMultiply(&(invH_best[k*k*i]),m_Vec1,m_Vec2,k,k,1);
			matrixMultiply(m_Vec1,m_Vec2,m_Vec3,1,k,1);
		}
		else
			matrixMultiply(m_Vec1,m_Vec1,m_Vec3,1,k,1);

		if(kType == EPANECHNIKOV)
			density += kEpanechnikov(*m_Vec3);
		else if(kType == TUKEY)
			density += kTukey(*m_Vec3);
		else if(kType == GAUSSIAN)
			density += kGaussian(*m_Vec3);
	}

	return -density;
}

double conjugateGradient::gmComputeFuncAndGrad(double *y, double *J)
{
	int i, j, kk;
	double density, dif, wt;

	// compute all projections
	innerProduct(y,theta_dim,Y,Y_dim,proj_all);

	// compute best projection per carrier
	computeBestCarriers(proj_all, proj_best, invH, &(y[k*n]), dists, ind_constraints, npts, nCarriers, k, isHeteroscedastic);
	
	// compute score and gradients
	memset(J, 0, sizeof(double) * p);
	if(isHeteroscedastic)
	{
		for(int pt_i = 0; pt_i < npts; pt_i++)
			memcpy(&(invH_best[pt_i*k*k]),&(invH[(pt_i*nCarriers + ind_constraints[pt_i])*k*k]),k*k*sizeof(double));
	}

	for(i = 0, density = 0, pProj = proj_best; i < npts; i++, pProj++)
	{
		double* pData = &(Y[n*nCarriers*i + n*ind_constraints[i]]);
		for(kk = 0; kk < k; kk++)
			m_Vec1[kk] = (pProj[kk*npts] - y[k*n+kk])/scale[kk];

		if(isHeteroscedastic)
		{
			matrixMultiply(&(invH_best[k*k*i]),m_Vec1,m_Vec2,k,k,1);
			matrixMultiply(m_Vec1,m_Vec2,m_Vec3,1,k,1);
		}
		else
		{
			memcpy(m_Vec2,m_Vec1,sizeof(double)*k);
			matrixMultiply(m_Vec1,m_Vec1,m_Vec3,1,k,1);
		}

		dif = 1 - *m_Vec3;

		if(dif < 0)
			continue;

		if(kType == EPANECHNIKOV)
		{
			density += kEpanechnikov(*m_Vec3);
			wt = 2;
		}
		else if(kType == TUKEY)
		{
			density += kTukey(*m_Vec3);
			wt = 2*dif*dif;
		}
		else if(kType == GAUSSIAN)
		{
			density += kGaussian(*m_Vec3);
			wt = 1;
		}

		for(kk = 0; kk < k; kk++){
			for(j = 0; j < n; j++){
				J[kk * n + j] -= wt * m_Vec2[kk] * pData[j] / scale[kk]; // corresponds to theta
			}
			if(!isHomogeneous)
				J[k * n + kk] += wt * m_Vec2[kk] / scale[kk];				// corresponds to alpha
		}
	}

	// get gradient for Grassmann manifold
	// m_dbArrW = y^t * J
	for(i = 0; i < k; i++){
		for(j = 0; j < k; j++){
			for(kk = 0, m_dbArrW[i * k + j] = 0; kk < n; kk++)
				m_dbArrW[i * k + j] += y[j * n + kk] * J[i * n + kk];
		}
	}
	// J = J - y * m_dbArrW
	for(i = 0; i < n; i++){
		for(j = 0; j < k; j++){
			for(kk = 0; kk < k; kk++){
				J[j * n + i] -= y[kk * n + i] * m_dbArrW[j * k + kk];
			}
		}
	}
	
	return -density;
}

void conjugateGradient::doGrassmannConjugateGradient(double ftol, int& iter)
{
	int its, i;
	double gg, gam, fp, dgg, cs, sn, dp;
	double fret;

	// values used in line minimization
	double cx, xmin, fc, fb, fa, bx, ax, tol = 2.0e-10;

	// Evaluation of G_{0} and H_{0}
	fp = gmComputeFuncAndGrad(m_ThetaAlpha, m_dbArrX);

	// initial gradient and direction of motion
	for(int i = 0; i < p; i++)
		m_G[i] = -m_dbArrX[i];

	memcpy(m_dbArrH, m_G, sizeof(double) * p);
	memcpy(m_dbArrX, m_G, sizeof(double) * p);

	// Compute norm of tangent vector
	gg = 0.0;
	for(i = 0, gg = 0.0; i < p; i++)
		gg += m_G[i] * m_G[i];

	// Unlikely to happen, but if gradient is exactly zero then we are done!!!
	if(gg < 1.0e-20){
		iter = 0;
		fret = fp;
		return;
	}

	for(its = 0; its < MAXITER; its++)
	{
		iter = its;
		// performing SVD, for line minimization
		dgesvd_("S", "A", &n, &k, m_dbArrX, &n, m_S, m_U, &n, m_V, &k, m_dbArrWork, &m_nLwork, &m_nInfo);

		//appending alpha in m_U (U)
		if(!isHomogeneous)
			memcpy(m_U + n * k, m_dbArrX + n * k, sizeof(double) * k);

		ax = 0.0;
		bx = 1.0;
		gmbrak(ax, bx, cx, fa, fb, fc);
		fret = gmbrent(ax, bx, cx, tol, xmin);

		if(fret > fp)
			return;

		// Compute norm of tangent vector for use later...
		if(!isHomogeneous)
		{
			for(i = 0, gg = 0; i < p; i++)
				gg += m_G[i] * m_G[i];
		}
		else
		{
			for(i = 0, gg = 0; i < n*k; i++)
				gg += m_G[i] * m_G[i];
		}

		// Unlikely to happen, but if gradient is exactly zero then we are done!!!
		if(gg < 1.0e-20)
			return;
		
		// parallel transport of G_{its}, H_{its}.
		for(i = 0; i < k; i++){
			m_Cos[i] = cos(m_S[i] * xmin);
			m_Sin[i] = sin(m_S[i] * xmin);
		}
		parallelTransport(m_ThetaAlpha, m_U, m_V, m_S, m_Cos, m_Sin, m_G, m_dbArrH);

		// get new value of Yk
		geodesicMove(m_ThetaAlpha, m_U, m_V, m_Cos, m_Sin, m_dbArrT);
		memcpy(m_ThetaAlpha, m_dbArrT, sizeof(double) * n * k);
		if(!isHomogeneous)
		{
			for(i = n * k; i < p; i++)
				m_ThetaAlpha[i] += m_dbArrX[i] * xmin;
		}

		// termination conditions
		if(2.0 * fabs(fret - fp) <= ftol * (fabs(fret) + fabs(fp) + TINY))
			return;

		// computation of G_{its+1}
		fp = gmComputeFuncAndGrad(m_ThetaAlpha, m_dbArrX);
		// computation of new search direction H_{its+1}

		dgg = 0;
		if(!isHomogeneous)
		{
			// Polak-Ribiere
			for(i = 0; i < p; i++)
				dgg += m_dbArrX[i] * (m_dbArrX[i] + m_G[i]);
			// Fletcer-Reeves
			//		for(i = 0; i < p; i++)
			//			dgg += m_dbArrX[i] * m_dbArrX[i];

			gam = dgg / gg;

			for(i = 0; i < p; i++){
				m_G[i] = -m_dbArrX[i];
				m_dbArrH[i] = m_G[i] + m_dbArrH[i] * gam;
			}
			memcpy(m_dbArrX, m_dbArrH, sizeof(double) * p);
		}
		else
		{
			// Polak-Ribiere
			for(i = 0; i < n*k; i++)
				dgg += m_dbArrX[i] * (m_dbArrX[i] + m_G[i]);
			// Fletcer-Reeves
			//		for(i = 0; i < p; i++)
			//			dgg += m_dbArrX[i] * m_dbArrX[i];

			gam = dgg / gg;

			for(i = 0; i < n*k; i++){
				m_G[i] = -m_dbArrX[i];
				m_dbArrH[i] = m_G[i] + m_dbArrH[i] * gam;
			}
			memcpy(m_dbArrX, m_dbArrH, sizeof(double) * n*k);
		}
	}
}


void conjugateGradient::localOptimize(double *theta, int& nIter)
{
	memcpy(m_ThetaAlpha, theta, p * sizeof(double));
	doGrassmannConjugateGradient(1.0e-6, nIter);
	memcpy(theta, m_ThetaAlpha, p * sizeof(double));
}
