#include <mex.h>
#include <matrix.h>
#include <string.h>
#include <time.h>
#include "commons.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray* prhs[])
{
	// Reading in all input variables
	double* Y = mxGetPr(prhs[0]);								// data matrix with carriers
	mwSize npts = mxGetN(prhs[0]);								// number of data points
	double* theta = mxGetPr(prhs[1]);							// estimated theta
	mwSize n = mxGetM(prhs[1]);									// dimensionality of the original space
	mwSize k = mxGetN(prhs[1]);									// dimensionality of null space
	double* C = mxGetPr(prhs[2]);								// covariance matrices
	double* mode = mxGetPr(prhs[3]);							// k-dimensional mode
	double* scale = mxGetPr(prhs[4]);							// k-dimensional scale 
	double* nCarriers_ = mxGetPr(prhs[5]);						// number of carriers per constraint
	int k_name_len = (mxGetM(prhs[6]) * mxGetN(prhs[6])) + 1;	// length of the kernel name
	double* isHeteroscedastic_ = mxGetPr(prhs[7]);				// flag if problem is heteroscedastic

	// convert input variables into proper types
	bool isHeteroscedastic = (*isHeteroscedastic_ != 0.0);
	int  nCarriers = (int)(*nCarriers_);						
	npts /= nCarriers;

	// Setting appropriate kernel
	char* kernel_name = (char*)mxMalloc(k_name_len*sizeof(char));
	mxGetString(prhs[6],kernel_name,k_name_len);
	kernelType ktype;
	if(!strcmp(kernel_name,"Epanechnikov"))
		ktype = EPANECHNIKOV;
	else if(!strcmp(kernel_name,"Gaussian"))
		ktype = GAUSSIAN;
	else if(!strcmp(kernel_name,"Tukey"))
		ktype = TUKEY;


	// Initializing variables
	double* proj_all = (double*)mxCalloc(k*npts*nCarriers,sizeof(double));			// projections of all carriers for all points
	double* proj_best = (double*)mxCalloc(k*npts,sizeof(double));					// projections of best carriers
	double* proj_i = (double*)mxMalloc(k*sizeof(double));							// projection of one point
	double* H = (double*)mxCalloc(k*k,sizeof(double));								// covariance matrices of the projections
	double* invH = (double*)mxCalloc(k*k*npts*nCarriers,sizeof(double));			// inverse of covariance of all projected point for all carriers
	double* invH_best = (double*)mxCalloc(k*k*npts,sizeof(double));					// set of best projection covariances for every point
	int* inliers_max = (int*)mxMalloc(npts*sizeof(int));							// maximum number of inlier points
	double* dists = (double*)mxMalloc(npts*sizeof(double));							// distances of points from projected hypothesis
	double* tempMatrix = (double*)mxMalloc(n*k*sizeof(double));						// temporary matrix to store theta^{T}*C
	double* tempVector = (double*)mxMalloc(k*sizeof(double));						// temporary vector to store proj^{T}*invH
	int* ind_constraints = (int*)mxMalloc(npts*sizeof(int));						// set of indices of best carrier for every point
	int nInliers = 0;																// number of inliers
	int theta_dim[2] = {n,k};														// size of theta
	int Y_dim[2] = {n,npts*nCarriers};												// size of data
	
	// compute all projections
	innerProduct(theta,theta_dim,Y,Y_dim,proj_all);

	// compute covariance matrices of projections and their inverse, H, invH
	if(isHeteroscedastic)
	{
		for(int pt_i = 0; pt_i < npts*nCarriers; pt_i++)
		{
			int tempMatrix_dim[2] = {n,k};
			matrixMultiply(&(C[pt_i*n*n]), theta, tempMatrix, n, n, k);
			innerProduct(theta,theta_dim,tempMatrix,tempMatrix_dim,H);
			matrixInverse(H,&invH[pt_i*k*k],k,k);
		}
	}

	// compute best carrier vector for every point
	computeBestCarriers(proj_all, proj_best, invH, mode, dists, ind_constraints, npts, nCarriers, k, isHeteroscedastic);

	if(isHeteroscedastic)
	{
		// copy covariance matrices of best carriers
		for(int pt_i = 0;pt_i < npts; pt_i++)
			memcpy(&(invH_best[pt_i*k*k]),&(invH[(pt_i*nCarriers + ind_constraints[pt_i])*k*k]),k*k*sizeof(double));

		// normalize every projected point with its covariance
		for(int pt_i = 0; pt_i < npts; pt_i++)
		{
			double norm_deviation = 0.0;
			double deviation = 0.0;
			for(int k_i = 0; k_i < k; k_i++)
			{
				proj_i[k_i] = proj_best[k*pt_i + k_i] - mode[k_i];
				norm_deviation += proj_i[k_i]*proj_i[k_i];
			}
			norm_deviation = sqrt(norm_deviation);
			matrixMultiply(proj_i,&(invH_best[k*k*pt_i]),tempVector,1,k,k);
			matrixMultiply(tempVector,proj_i,&deviation,1,k,1);
			deviation = sqrt(deviation);
			for(int k_i = 0; k_i < k; k_i++)
				proj_best[k*pt_i + k_i] = mode[k_i] + proj_i[k_i]*deviation/norm_deviation;
		}
	}

	// compute norm of the scale
	double norm_scale = 0.0;
	for(int i = 0; i < k; i++)
		norm_scale += scale[i]*scale[i];
	
	norm_scale = sqrt(norm_scale);

	// run mean shift iterations from every projected point to do inlier/outlier dichotomy
	double* mshift;
	double norm_mshift = 100000.00;
	for(int i = 0; i < npts; i++)
	{
		int nMShiftIter = 0;
		memcpy(proj_i, proj_best + i*k, k*sizeof(double));
		do
		{
 			mshift = meanShift_cpp(proj_best, invH_best, proj_i, k, scale, npts, ktype, 0);
			nMShiftIter++;
			norm_mshift = 0.0;
			for(int j = 0; j < k; j++)
			{
				proj_i[j] += mshift[j];
				norm_mshift += mshift[j]*mshift[j]; 
			}
			mxFree(mshift);
			norm_mshift = sqrt(norm_mshift);
		}while(norm_mshift > 0.01 && nMShiftIter < 10);
		
		double norm_diff = 0.0;

		for(int j = 0; j < k; j++)
			norm_diff += (proj_i[j] - mode[j])*(proj_i[j] - mode[j]);
		norm_diff = sqrt(norm_diff);
		
		if(norm_diff <= norm_scale)
			inliers_max[nInliers++] = i+1;
	}

	// copy to output
	double* inliers = (double*)mxMalloc(nInliers*sizeof(double));
	for(int i = 0; i < nInliers; i++)
		inliers[i] = (double)inliers_max[i];

	mxArray* mxInliers = mxCreateDoubleMatrix(1,nInliers,mxREAL);
	plhs[0] = mxInliers;
	memcpy(mxGetPr(mxInliers),inliers,nInliers*sizeof(double));

	// deallocate memory
	if(nInliers > 0)
		mxFree(inliers);
	mxFree(proj_best);
	mxFree(proj_all);
	mxFree(proj_i);
	mxFree(H);
	mxFree(invH);
	if(npts > 0)
		mxFree(inliers_max);
	mxFree(dists);
	mxFree(tempMatrix);
	mxFree(tempVector);
	mxFree(ind_constraints);
	mxFree(invH_best);
}
