#include "commons.h"

using namespace std;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray* prhs[])
{
	// Reading in all input variables
	double* Y = mxGetPr(prhs[0]);							// data matrix with carriers
	mwSize npts_ = mxGetN(prhs[0]);							// number of points
	double* n_ = mxGetPr(prhs[1]);							// dimensionality of data
	double* k_ = mxGetPr(prhs[2]);							// dimensionality of the null-space
	double* hypsInS = mxGetPr(prhs[3]);						// set of elemental hypotheses
	mwSize nIter_ = mxGetN(prhs[3]);						// number of iterations (M)
	double* nCarriers_ = mxGetPr(prhs[4]);					// number of carriers per point
	double* C = mxGetPr(prhs[5]);							// set of carrier covariance matrices
	double* N_ = mxGetPr(prhs[6]);							// size of elemental subset
	double* isHomogeneous_ = mxGetPr(prhs[7]);				// flag if problem is homogeneous (Ax = 0)
	double* EPSILON_ = mxGetPr(prhs[8]);					// epsilon
	double* isHeteroscedastic_ = mxGetPr(prhs[9]);			// flag if problem is heteroscedastic

	// convert input variables into proper types
	int N = (int)*N_;									
	bool isHomogeneous = *isHomogeneous_ != 0.0;
	bool isHeteroscedastic = *isHeteroscedastic_ != 0.0;
	int nCarriers = (int)*nCarriers_;
	int npts = npts_/nCarriers;
	double EPSILON = (double)*EPSILON_;
	int n = (int)*n_;		
	int k = (int)*k_;		
	int nIter = (int)nIter_;
	

	// initializing variables
	int nFracs = 40;															// number of different fractions to be tried between [0,1]
	double* theta;																// theta estimated from elemental subset
	int theta_dim[2] = {n,k};													// size of theta
	int data_dim[2] = {n,N*nCarriers};											// size of elemental subset data matrix
	int Y_dim[2] = {n,npts*nCarriers};											// size of carrier data matrix
	double* frac = (double*)mxMalloc(nFracs*sizeof(double));					// fraction values between [0,1]
	int* hyp = (int*)mxMalloc(N*sizeof(int));									// elemental subset hypotheses point indices
	double* data = (double*)mxMalloc(n*N*nCarriers*sizeof(double));				// elemental subset data
	double* meanData = (double*)mxMalloc(n*sizeof(double));						// mean of elemental subset data
	double* centeredData = (double*)mxMalloc(n*N*nCarriers*sizeof(double));		// centered elemental subset data
	double* proj_best = (double*)mxMalloc(k*npts*sizeof(double));				// projections of best carriers
	double* proj_all = (double*)mxMalloc(k*npts*nCarriers*sizeof(double));		// projections of all carriers
	double* alpha = (double*)mxCalloc(k,sizeof(double));						// projection of hypothesis points (all at one point)
	double* volume = (double*)mxMalloc(nIter*nFracs*sizeof(double));			// minimum volume for all iterations and all fractions (M x nFracs)
	double* H = (double*)mxMalloc(k*k*sizeof(double));							// covariance matrix of a projected point
	double* invH = (double*)mxMalloc(k*k*npts*nCarriers*sizeof(double));		// inverse of covariance of a projected point
	double* dists_sorted_cum = (double*)mxMalloc(npts*sizeof(double));			// cumulative distances (sorted) of points from projected hypothesis
	int* idx_sorted = (int*)mxMalloc(npts*sizeof(int));							// set of sorted indices of all points (variable for quick sort)				
	double* tempMatrix = (double*)mxMalloc(n*k*sizeof(double));					// temporary matrix to store theta^{T}*C
	int* ind_constraints = (int*)mxMalloc(npts*sizeof(int));					// set of indices of best carrier for every point

	// variables for computing svd
	int Info;														
	int Lwork = 5*max(n,N*nCarriers);
	double* U = (double*)mxMalloc(n*n*sizeof(double));
	double* V = (double*)mxMalloc(N*nCarriers*N*nCarriers*sizeof(double));
	double* S = (double*)mxMalloc(min(n,N*nCarriers)*sizeof(double));
	double* Work = (double*)mxMalloc(Lwork*sizeof(double));

	// setting fractions
	double fracStep = 1.0/((double)nFracs);
	frac[0] = fracStep;
	for(int i = 1; i < nFracs; i++)
		frac[i] = frac[i-1] + fracStep;

	// initializing volume
	for(int i = 0; i < nIter*nFracs; i++)
		volume[i] = 1000000000.00;
	
	// loop over all M iterations
	for(int i = 0; i < nIter; i++)
	{
		// reading hypothesis indices
		for(int hyp_i = 0; hyp_i < N; hyp_i++)
			hyp[hyp_i] = (int)hypsInS[i*N + hyp_i];

		// populating data with the samples selected as hypotheses in this iteration
		int dataIndex = 0;
		for(int hyp_i = 0; hyp_i < N; hyp_i++)
			for(int row_i = 0; row_i < n*nCarriers; row_i++)
				data[dataIndex++] = Y[(hyp[hyp_i]-1)*n*nCarriers + row_i];
		
		// computing the mean of the data
		dataIndex = 0;		
		for(int dim_i = 0; dim_i < n; dim_i++)
		{	
			double sum = 0.0;
			for(int hyp_i = 0; hyp_i < N*nCarriers; hyp_i++)
			{				
				sum += data[hyp_i * n + dim_i];
				dataIndex += n;
			}
			meanData[dim_i] = (double)sum/(double)(N*nCarriers);
		}

		// subtract mean if solving an inhomogeneous system of equations
		if(!isHomogeneous) //(N*nCarriers == (n-k+1))
		{
			// centering the data
			for(int dim_i = 0; dim_i < n; dim_i++)
				for(int hyp_i = 0; hyp_i < N*nCarriers; hyp_i++)
					centeredData[hyp_i * n + dim_i] = data[hyp_i * n + dim_i] - meanData[dim_i];
		}
		else // not subtracting the mean otherwise
		{
			for(int dim_i = 0; dim_i < n; dim_i++)
				for(int hyp_i = 0; hyp_i < N*nCarriers; hyp_i++)
					centeredData[hyp_i * n + dim_i] = data[hyp_i * n + dim_i];
		}
		
		// take svd to get theta from elemental subset
		int N_nCarriers = nCarriers*N;
		dgesvd_("A", "N", &n, &N_nCarriers, centeredData, &n, S, U, &n, V, &N_nCarriers, Work, &Lwork, &Info);

		// extracting theta - last k cols from U
		theta = U + (n-k)*n;
		
		// compute all projections
		innerProduct(theta,theta_dim,Y,Y_dim,proj_all);

		// compute mean of projections of the hypothesis
		for(int dim_i = 0; dim_i < k; dim_i++)
		{
			alpha[dim_i] = 0.0;
			for(int hyp_i=0; hyp_i < N; hyp_i++)
				for(int const_i = 0; const_i < nCarriers; const_i++)
					alpha[dim_i] += proj_all[((hyp[hyp_i] - 1)*nCarriers + const_i)*k + dim_i];

			alpha[dim_i] /= (double)(N*nCarriers);
		}
		
		// compute covariance matrices of projections and their inverse, H, invH
		if(isHeteroscedastic)
		{
			for(int pt_i = 0; pt_i < npts*nCarriers; pt_i++)
			{
				int tempMatrix_dim[2] = {n,k};
				matrixMultiply(&(C[pt_i*n*n]), theta, tempMatrix, n, n, k);
				innerProduct(theta,theta_dim,tempMatrix,tempMatrix_dim,H);
				matrixInverse(H,&invH[pt_i*k*k],k,k);
			}
		}

		// compute best carrier vector for every point
		computeBestCarriers(proj_all, proj_best, invH, alpha, dists_sorted_cum, ind_constraints, npts, nCarriers, k, isHeteroscedastic);

		// initialize indices before sorting
		for(int pt_i = 0; pt_i < npts; pt_i++)
			idx_sorted[pt_i] = pt_i;

		// sort points in ascending distances from projected hypothesis
		quickSort(dists_sorted_cum, idx_sorted, 0, npts-1);

		// compute cumulative distances
		for(int pt_i = 1; pt_i < npts; pt_i++)
			dists_sorted_cum[pt_i] += dists_sorted_cum[pt_i-1]; 
		
		// compute volume
		for(int fracIter = 0; fracIter < nFracs; fracIter++)
		{
			int ndense = (int)(npts*frac[fracIter]);
			if(ndense >= npts)
				ndense = npts-1;
			if(ndense == 0)
				volume[fracIter*nIter + i] = 0.0;
			else
			{
				double detScale = sqrt(dists_sorted_cum[ndense]);
				volume[fracIter*nIter + i] = (ndense-N)/(detScale + EPSILON);
			}
		}
	}
	
	// copy to output
	const int dims_theta_all[3] = {n,k,nFracs};
	mxArray* mxNorm_minDist = mxCreateDoubleMatrix(nIter,nFracs,mxREAL);
	plhs[0] = mxNorm_minDist;
	memcpy(mxGetPr(mxNorm_minDist),volume,nIter*nFracs*sizeof(double));
	
	// deallocate memory
	mxFree(frac);
	mxFree(hyp);
	mxFree(data);
	mxFree(meanData);
	mxFree(centeredData);
	mxFree(U);
	mxFree(V);
	mxFree(S);
	mxFree(Work);
	mxFree(proj_best);
	mxFree(volume);
	mxFree(proj_all);
	mxFree(alpha);
	mxFree(H);
	mxFree(invH);
	mxFree(dists_sorted_cum);
	mxFree(idx_sorted);
	mxFree(tempMatrix);
	mxFree(ind_constraints);
}

