#include "commons.h"
#include "conjugateGradient.h"

void getScore(double* projections, int npts, double* theta, double* alpha, int n, int k, 
			  double *invH, double *scale, double &score, int &nInliers, double* &inliers, 
			  int &nMShiftIter, double *mode, kernelType ktype, bool isHeteroscedastic);
void doConjugateGradient(double& score, double* scale, double* theta, int theta_dim[2], double* mode, 
						 double* Y, int Y_dim[2], double* C, int nCarriers, int& nInliers, double* &inliers, 
						 int n, int k, int N, int npts, double* invH, kernelType ktype, bool isHeteroscedastic, bool isHomogeneous);
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray* prhs[])
{
	// Reading in all input variables
	double* Y = mxGetPr(prhs[0]);									// data matrix with carriers
	mwSize npts_ = mxGetN(prhs[0]);									// number of points
	double* n_ = mxGetPr(prhs[1]);									// dimensionality of original space
	double* k_ = mxGetPr(prhs[2]);									// dimensionality of null space
	double* hypsIn = mxGetPr(prhs[3]);								// elemental hypotheses
	mwSize nIter_ = mxGetN(prhs[3]);								// number of iterations (N)
	double* C = mxGetPr(prhs[4]);									// npts covariance matrices
	double* scale = mxGetPr(prhs[5]);								// estimated scale
	int k_name_len = (mxGetM(prhs[6]) * mxGetN(prhs[6])) + 1;		// length of the kernel type name
	double* nCarriers_ = mxGetPr(prhs[7]);							// number of carriers per constraint
	double* N_ = mxGetPr(prhs[8]);									// size of elemental subset
	double* isHomogeneous_ = mxGetPr(prhs[9]);						// flag if problem is homogeneous (Ax = b)
	double* isGrassmann_ = mxGetPr(prhs[10]);						// flag if local optimization on Grassmann manifolds should be performed or not
	double* isHeteroscedastic_ = mxGetPr(prhs[11]);					// flag if problem is heteroscedastic

	// convert input variables into proper types
	int N = (int)*N_;												
	bool isHomogeneous = (*isHomogeneous_ != 0.0);
	bool isGrassmann = (*isGrassmann_ != 0.0);
	bool isHeteroscedastic = (*isHeteroscedastic_ != 0.0);
	int nCarriers = (int)*nCarriers_;
	int n = (int)*n_;												// dimension of the problem
	int k = (int)*k_;												// dimension of the null space
	int nIter = (int)nIter_;
	int npts = npts_/nCarriers;

	//select appropriate kernel
	char* kernel_name = (char*)mxMalloc(k_name_len*sizeof(char));
	mxGetString(prhs[6],kernel_name,k_name_len);
	kernelType ktype;
	if(!strcmp(kernel_name,"Epanechnikov"))
		ktype = EPANECHNIKOV;
	else if(!strcmp(kernel_name,"Gaussian"))
		ktype = GAUSSIAN;
	else if(!strcmp(kernel_name,"Tukey"))
		ktype = TUKEY;

	// initialize best score
	double scoreBest = -100.00;

	// initializing variables
	double* theta;																	// theta estimated from elemental subset
	double* alpha = (double*)mxCalloc(k, sizeof(double));							// alpha estimated from elemental subset
	double* mode = (double*)mxCalloc(k,sizeof(double));								// mode estimated using mean shift
	int theta_dim[2] = {n,k};														// size of theta
	int data_dim[2] = {n,N*nCarriers};												// size of elemental subset data matrix
	int Y_dim[2] = {n,npts*nCarriers};												// size of carrier data matrix
	double* data = (double*)mxCalloc(n*N*nCarriers, sizeof(double));				// elemental subset data
	double* meanData = (double*)mxCalloc(n,sizeof(double));							// mean of elemental subset data
	double* centeredData = (double*)mxCalloc(n*N*nCarriers,sizeof(double));			// centered elemental subset data
	double* proj_best = (double*)mxMalloc(k*npts*sizeof(double));					// projections of best carriers
	double* proj_all = (double*)mxMalloc(k*npts*nCarriers*sizeof(double));			// projections of all carriers
	int* hyp = (int*)mxMalloc(N*sizeof(int));										// elemental subset hypotheses point indices
	double* hyp_double = (double*)mxMalloc(N*sizeof(double));						// output best hypothesis
	double* H = (double*)mxMalloc(k*k*sizeof(double));								// covariance matrix of a projected point
	double* invH = (double*)mxCalloc(k*k*npts*nCarriers,sizeof(double));			// inverse of covariance of all projected point for all carriers
	double* dists = (double*)mxMalloc(npts*sizeof(double));							// distances of points from projected hypothesis
	double* tempMatrix = (double*)mxMalloc(n*k*sizeof(double));						// temporary matrix to store theta^{T}*C
	int* ind_constraints = (int*)mxMalloc(npts*sizeof(int));						// set of indices of best carrier for every point
	double* invH_best = (double*)mxCalloc(k*k*npts,sizeof(double));					// set of best projection covariances for every point

	// variables for computing svd
	int Info;
	int Lwork = 5*max(n,N*nCarriers);
	double* Work = (double*)mxMalloc(Lwork*sizeof(double));
	double* U = (double*)mxMalloc(n*n*sizeof(double));
	double* V = (double*)mxMalloc(N*nCarriers*N*nCarriers*sizeof(double));
	double* S = (double*)mxMalloc(min(n,N*nCarriers)*sizeof(double));

	// initialize output variables
	mxArray* mxTheta = mxCreateDoubleMatrix((mwSize)n,(mwSize)k,mxREAL);
	mxArray* mxAlpha = mxCreateDoubleMatrix((mwSize)k,1,mxREAL);
	mxArray* mxScore = mxCreateDoubleMatrix(1,1,mxREAL);
	mxArray* mxInliers;
	mxArray* mxBestHyp = mxCreateDoubleMatrix(N,1,mxREAL);

	int nMShiftIter;
	double *inliers;
	int nInliers;
	double score = 0.0;

	for(int i = 0; i < nIter; ++i)
	{
		for(int hyp_i = 0; hyp_i < N; hyp_i++)
			hyp[hyp_i] = (int)hypsIn[i*N + hyp_i];

		// populating data with the samples selected as hypotheses in this iteration
		int dataIndex = 0;
		for(int hyp_i = 0; hyp_i < N; hyp_i++)
			for(int row_i = 0; row_i < n*nCarriers; row_i++)
				data[dataIndex++] = Y[(hyp[hyp_i]-1)*n*nCarriers + row_i];

		// computing the mean of the data
		dataIndex = 0;		
		for(int dim_i = 0; dim_i < n; dim_i++)
		{	
			double sum = 0.0;
			for(int hyp_i = 0; hyp_i < N*nCarriers; hyp_i++)
			{				
				sum += data[hyp_i * n + dim_i];
				dataIndex += n;
			}
			meanData[dim_i] = (double)sum/(double)(N*nCarriers);
		}

		// subtract mean if solving an inhomogeneous system of equations
		if(!isHomogeneous) // (N*nCarriers == n-k+1)
		{
			// centering the data
			for(int dim_i = 0; dim_i < n; dim_i++)
				for(int hyp_i = 0; hyp_i < N*nCarriers; hyp_i++)
					centeredData[hyp_i * n + dim_i] = data[hyp_i * n + dim_i] - meanData[dim_i];
		}
		else // not subtracting the mean otherwise
		{
			for(int dim_i = 0; dim_i < n; dim_i++)
				for(int hyp_i = 0; hyp_i < N*nCarriers; hyp_i++)
					centeredData[hyp_i * n + dim_i] = data[hyp_i * n + dim_i];
		}

		// take svd to get theta from elemental subset
		int N_nCarriers = nCarriers*N;
		dgesvd_("A", "N", &n, &N_nCarriers, centeredData, &n, S, U, &n, V, &N_nCarriers, Work, &Lwork, &Info);

		// extracting theta - last k cols from U
		theta = U + (n-k)*n;

		// compute all projections
		innerProduct(theta,theta_dim,Y,Y_dim,proj_all);

		// compute mean of projections of the hypothesis
		for(int dim_i = 0; dim_i < k; dim_i++)
		{
			alpha[dim_i] = 0.0;
			for(int hyp_i=0; hyp_i < N; hyp_i++)
				for(int const_i = 0; const_i < nCarriers; const_i++)
					alpha[dim_i] += proj_all[((hyp[hyp_i] - 1)*nCarriers + const_i)*k + dim_i];

			alpha[dim_i] /= (double)(N*nCarriers);
		}

		// compute covariance matrices of projections and their inverse, H, invH
		if(isHeteroscedastic)
		{
			for(int pt_i = 0; pt_i < npts*nCarriers; pt_i++)
			{
				int tempMatrix_dim[2] = {n,k};
				matrixMultiply(&(C[pt_i*n*n]), theta, tempMatrix, n, n, k);
				innerProduct(theta,theta_dim,tempMatrix,tempMatrix_dim,H);
				matrixInverse(H,&invH[pt_i*k*k],k,k);
			}
		}

		// compute best carrier vector for every point
		computeBestCarriers(proj_all, proj_best, invH, alpha, dists, ind_constraints, npts, nCarriers, k, isHeteroscedastic);

		// copy covariance matrices of best carriers
		if(isHeteroscedastic)
		{
			for(int pt_i =0;pt_i < npts;pt_i++)
				memcpy(&(invH_best[pt_i*k*k]),&(invH[(pt_i*nCarriers + ind_constraints[pt_i])*k*k]),k*k*sizeof(double));
		}

		// compute score
		getScore(proj_best, npts, theta, alpha, n, k, invH_best, scale, score, nInliers, inliers, nMShiftIter, mode, ktype, isHeteroscedastic);
		
		// do conjugate gradient on Grassmann manifold
		if(isGrassmann && score > 0.9*scoreBest)
			doConjugateGradient(score, scale, theta, theta_dim, mode, Y, Y_dim, C, nCarriers, nInliers, inliers, n, k, N, npts, invH, ktype, isHeteroscedastic, isHomogeneous);

		// if current score is best so far, copy all estimated variables to output
		if(score > scoreBest)
		{	
			scoreBest = score;

			memcpy(mxGetPr(mxTheta),theta,n*k*sizeof(double));
			plhs[0] = mxTheta;

			memcpy(mxGetPr(mxAlpha),mode,k*sizeof(double));
			plhs[1] = mxAlpha;

			memcpy(mxGetPr(mxScore),&score,sizeof(double));
			plhs[2] = mxScore;

			if(i > 0)
				mxDestroyArray(mxInliers);
			mxInliers = mxCreateDoubleMatrix(1,nInliers,mxREAL);
			memcpy(mxGetPr(mxInliers),inliers,nInliers*sizeof(double));

			plhs[3] = mxInliers;
			for(int hyp_i = 0; hyp_i < N; hyp_i++)
				hyp_double[hyp_i] = (double)hyp[hyp_i];

			memcpy(mxGetPr(mxBestHyp),hyp_double,N*sizeof(double));
			plhs[4] = mxBestHyp;
		}
		if(nInliers > 0)
			mxFree(inliers);
	}
	
	// deallocate memory
	mxFree(mode);
	mxFree(alpha);
	mxFree(data);
	mxFree(meanData);
	mxFree(centeredData);
	mxFree(U);
	mxFree(V);
	mxFree(S);
	mxFree(Work);
	mxFree(kernel_name);
	mxFree(proj_best);
	mxFree(proj_all);
	mxFree(hyp);
	mxFree(hyp_double);
	mxFree(invH);
	mxFree(H);
	mxFree(dists);
	mxFree(tempMatrix);
	mxFree(ind_constraints);
	mxFree(invH_best);
	return;
}

void getScore(double* projections, int npts, double* theta, double* alpha, int n, int k, double *invH, double *scale, double &score, int &nInliers, double* &inliers, int &nMShiftIter, double *mode, kernelType ktype, bool isHeteroscedastic)
{
	// initialize variables
	nMShiftIter = 0;				// number of mean shift iterations
	double* mshift;					// mean shift vector
	double norm_mshift = 100.00;

	// initialize mode as alpha (projection of hypothesis points)
	for(int i = 0; i < k; i++)
		mode[i] = alpha[i];	

	// perform mean shift iterations until convergence
	do
	{
		mshift = meanShift_cpp(projections, invH, mode, k, scale, npts, ktype, isHeteroscedastic);
		nMShiftIter++;
		norm_mshift = 0.0;
		for(int i = 0; i < k; i++)
		{
			mode[i] += mshift[i];
			norm_mshift += mshift[i]*mshift[i]; 
		}
		mxFree(mshift);
		norm_mshift = sqrt(norm_mshift);
	}while(norm_mshift > 0.00001 && nMShiftIter < 100);

	//initialize temporary variables
	double *invHi, *projectionsi;
	double *tempMul2, *tempMul3, *tempMul4;
	int *inliers_ = (int*)mxMalloc(npts*sizeof(int));
	nInliers = 0;
	invHi = (double*)mxMalloc(k*k*sizeof(double));
	projectionsi = (double*)mxMalloc(k*sizeof(double));
	tempMul2 = (double*)mxMalloc(k*sizeof(double));
	tempMul3 = (double*)mxMalloc(k*sizeof(double));
	tempMul4 = (double*)mxMalloc(1*sizeof(double));

	// compute score
	score = 0.0;
	for(int i = 0; i < npts; i++)
	{
		// compute heteroscedastic deviations
		memcpy(projectionsi, projections + i*k, k*sizeof(double));
		for(int j = 0; j < k; j++)
			tempMul2[j] = (mode[j] - projectionsi[j])/(scale[j]); 

		double det = 1.0;
		if(isHeteroscedastic)
		{
			memcpy(invHi, invH + i*k*k, k*k*sizeof(double));
			matrixMultiply(tempMul2,invHi,tempMul3,1,k,k);
			matrixMultiply(tempMul3,tempMul2,tempMul4,1,k,1);
			det = matrixDeterminant(invHi,k);
			det = sqrt(det);
		}
		else
			matrixMultiply(tempMul2,tempMul2,tempMul4,1,k,1);

		// evaluating the kernel to compute score and decide if the i-th point is inlier or not
		if(ktype == EPANECHNIKOV)
		{
			score += kEpanechnikov(*tempMul4)*det;
			if(kEpanechnikov(*tempMul4) > 0)
				inliers_[nInliers++] = i+1;
		}
		else if(ktype == GAUSSIAN)
		{
			score += kGaussian(*tempMul4)*det;
			if(kGaussian(*tempMul4) > 0)
				inliers_[nInliers++] = i+1;
		}
		else if(ktype == TUKEY)
		{
			score += kTukey(*tempMul4)*det;
			if(kTukey(*tempMul4) > 0)
				inliers_[nInliers++] = i+1;
		}
	}

	// copy inliers
	inliers = (double*)mxMalloc(nInliers*sizeof(double));
	for(int i = 0; i < nInliers; i++)
		inliers[i] = (double)inliers_[i];

	// normalize score
	score = score/((double)npts);

	// deallocate memory
	if(npts > 0)
		mxFree(inliers_);
	mxFree(projectionsi);
	mxFree(invHi);
	mxFree(tempMul2);
	mxFree(tempMul3);
	mxFree(tempMul4);
}

void doConjugateGradient(double& score, double* scale, double* theta, int theta_dim[2], double* mode, 
						 double* Y, int Y_dim[2], double* C, int nCarriers, int& nInliers, double* &inliers, 
						 int n, int k, int N, int npts, double* invH, kernelType ktype, bool isHeteroscedastic, bool isHomogeneous)
{
	double* H = (double*)mxMalloc(k*k*sizeof(double));								// covariance matrix of a projected point
	double* invH_best = (double*)mxCalloc(k*k*npts,sizeof(double));					// set of best projection covariances for every point
	double* dists = (double*)mxMalloc(npts*sizeof(double));							// cumulative distances (sorted) of points from projected hypothesis
	double* tempMatrix = (double*)mxMalloc(n*k*sizeof(double));						// temporary matrix to store theta^{T}*C
	int* ind_constraints = (int*)mxMalloc(npts*sizeof(int));						// set of indices of best carrier for every point
	double* proj_best = (double*)mxMalloc(k*npts*sizeof(double));					// projections of best carriers
	double* proj_all = (double*)mxMalloc(k*npts*nCarriers*sizeof(double));			// projections of all carriers
	double* thetaTemp = (double*)mxMalloc(sizeof(double)*n*k);
	double* alphaTemp = (double*)mxMalloc(sizeof(double)*k);
	double* thetaAlpha = (double*)mxMalloc(sizeof(double)*(n+1)*k);
	double* modeTemp = (double*)mxMalloc(sizeof(double)*k);

	int nMShiftIter;
	
	// store score computed before CG
	double scoreBefore = score;

	// store thetaAlpha before CG
	memcpy(thetaAlpha,theta,sizeof(double)*n*k);

	if(!isHomogeneous)
	{
		for(int cg_i = 0; cg_i < k; cg_i++)
			thetaAlpha[n*k + cg_i] = mode[cg_i];
	}
	else
	{
		for(int cg_i = 0; cg_i < k; cg_i++)
			thetaAlpha[n*k + cg_i] = 0.0;
	}

	// perform CG
	conjugateGradient cgGpbM;
	cgGpbM.initializeData(Y,n,N,nCarriers,npts,k,scale,invH,ktype,isHeteroscedastic,isHomogeneous);
	int nCgIter;
	cgGpbM.localOptimize(thetaAlpha,nCgIter);
	mexPrintf("Number of Grassmann iterations = %d\n",nCgIter);

	// optimized theta, alpha and set of inliers
	double* inliersTemp;
	int nInliersTemp = 0;
	memcpy(thetaTemp,thetaAlpha,sizeof(double)*n*k);
	for(int cg_i = 0; cg_i < k; cg_i++)
		alphaTemp[cg_i] = thetaAlpha[n*k + cg_i];

	// compute all projections
	innerProduct(thetaTemp,theta_dim,Y,Y_dim,proj_all);

	// compute covariance matrices of proj_best, H
	if(isHeteroscedastic)
	{
		for(int pt_i = 0; pt_i < npts*nCarriers; pt_i++)
		{
			int tempMatrix_dim[2] = {n,k};
			matrixMultiply(&(C[pt_i*n*n]), thetaTemp, tempMatrix, n, n, k);
			innerProduct(thetaTemp,theta_dim,tempMatrix,tempMatrix_dim,H);
			matrixInverse(H,&invH[pt_i*k*k],k,k);
		}
	}

	// compute best carrier vector for every point
	computeBestCarriers(proj_all, proj_best, invH, alphaTemp, dists, ind_constraints, npts, nCarriers, k, isHeteroscedastic);

	// copy covariance matrices of best carriers
	if(isHeteroscedastic)
	{
		for(int pt_i =0;pt_i < npts;pt_i++)
			memcpy(&(invH_best[pt_i*k*k]),&(invH[(pt_i*nCarriers + ind_constraints[pt_i])*k*k]),k*k*sizeof(double));
	}
	// compute score
	getScore(proj_best, npts, thetaTemp, alphaTemp, n, k, invH_best, scale, score, nInliersTemp, inliersTemp, nMShiftIter, modeTemp, ktype,isHeteroscedastic);
	// use this score if it is better than the one computed before CG
	mexPrintf("scoreBefore = %f, nInliersBefore = %d, scoreAfter = %f, nInliersAfter = %d\n",scoreBefore,nInliers,score,nInliersTemp);

	if(score > scoreBefore)
	{
		memcpy(theta,thetaAlpha,sizeof(double)*n*k);

		for(int cg_i = 0; cg_i < k; cg_i++)
			mode[cg_i] = modeTemp[cg_i];
		if(nInliers > 0)
			mxFree(inliers);

		nInliers = nInliersTemp;
		inliers = (double*)mxMalloc(nInliers*sizeof(double));
		memcpy(inliers,inliersTemp,nInliers*sizeof(double));
	}
	else // otherwise use the score, theta, alpha etc. before CG
		score = scoreBefore;

	// clear variables used in CG
	cgGpbM.clearData();

	mxFree(thetaAlpha);
	mxFree(thetaTemp);
	mxFree(alphaTemp);
	mxFree(H);
	mxFree(invH_best);
	mxFree(dists);
	mxFree(tempMatrix);
	mxFree(ind_constraints);
	mxFree(proj_best);
	mxFree(proj_all);
	if(nInliersTemp > 0)
		mxFree(inliersTemp);

	mxFree(modeTemp);

}