#include "commons.h"

using namespace std;

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray* prhs[])
{
	// Reading in all input variables
	double* Y = mxGetPr(prhs[0]);						// data matrix with carriers
	mwSize npts_ = mxGetN(prhs[0]);						// number of points
	double* n_ = mxGetPr(prhs[1]);						// dimensionality of data
	double* k_ = mxGetPr(prhs[2]);						// dimensionality of the null-space
	double* frac_ = mxGetPr(prhs[3]);					// estimated fraction of inliers
	double* hypsIn = mxGetPr(prhs[4]);					// set of estimated hypotheses
	double* nCarriers_ = mxGetPr(prhs[5]);				// number of carrier per point
	double* C = mxGetPr(prhs[6]);						// set of carrier covariance matrices
	double* N_ = mxGetPr(prhs[7]);						// size of elemental subset
	double* isHomogeneous_ = mxGetPr(prhs[8]);			// flag if problem is homogeneous (Ax = 0)
	double* isHeteroscedastic_ = mxGetPr(prhs[9]);		// flag if problem is heteroscedastic


	// convert input variables into proper types
	int N = (int)*N_;
	bool isHomogeneous = *isHomogeneous_ != 0.0;
	bool isHeteroscedastic = *isHeteroscedastic_ != 0.0;
	int n = (int)*n_;
	int k = (int)*k_;
	double frac = (double)*frac_;
	int nCarriers = (int)*nCarriers_;
	int npts = npts_/nCarriers;
	
	double* theta;															// theta estimated from elemental subset
	int theta_dim[2] = {n,k};												// size of theta
	int data_dim[2] = {n,N*nCarriers};										// size of elemental subset data matrix
	int Y_dim[2] = {n,npts*nCarriers};										// size of carrier data matrix
	int ndense = (int)(npts*frac);											// number of points in the estimated fraction
	int* hyp = (int*)mxMalloc(N*sizeof(int));								// elemental subset hypothesis
	double* data = (double*)mxMalloc(n*N*nCarriers*sizeof(double));			// elemental subset data
	double* meanData = (double*)mxMalloc(n*sizeof(double));					// mean of elemental subset data
	double* centeredData = (double*)mxMalloc(n*N*nCarriers*sizeof(double));	// centered elemental subset data
	double* proj_best = (double*)mxMalloc(k*npts*sizeof(double));			// projections of best carriers
	double* proj_all = (double*)mxMalloc(k*npts*nCarriers*sizeof(double));	// projections of all carriers
	int* inliersS = (int*)mxMalloc(ndense*sizeof(int));						// inliers returned	by scale estimation step
	double* alpha = (double*)mxCalloc(k,sizeof(double));					// projection of the hypothesis points (all at one point)
	double* H = (double*)mxMalloc(k*k*sizeof(double));						// covariance matrix of a projected point
	double* invH = (double*)mxMalloc(k*k*npts*nCarriers*sizeof(double));	// inverse of covariance of a projected point
	double* dists_sorted = (double*)mxMalloc(npts*sizeof(double));			// sorted distances of points from projected hypothesis
	int* idx_sorted = (int*)mxMalloc(npts*sizeof(int));						// set of sorted indices of all points (variable for quick sort)
	double* tempMatrix = (double*)mxMalloc(n*k*sizeof(double));				// temporary matrix to store theta^{T}*C
	int* ind_constraints = (int*)mxMalloc(npts*sizeof(int));				// set of indices of best carrier for every point

	// variables for computing svd
	int Info;
	int Lwork = 5*max(n,N*nCarriers);
	double* Work = (double*)mxMalloc(Lwork*sizeof(double));
	double* U = (double*)mxMalloc(n*n*sizeof(double));
	double* V = (double*)mxMalloc(N*nCarriers*N*nCarriers*sizeof(double));
	double* S = (double*)mxMalloc(min(n,N*nCarriers)*sizeof(double));

	// reading hypothesis indices
	for(int hyp_i = 0; hyp_i < N; hyp_i++)
		hyp[hyp_i] = (int)hypsIn[hyp_i];

	// populating data with the samples selected as hypothesis
	int dataIndex = 0;
	for(int hyp_i = 0; hyp_i < N; hyp_i++)
		for(int row_i = 0; row_i < n*nCarriers; row_i++)
			data[dataIndex++] = Y[(hyp[hyp_i]-1)*n*nCarriers + row_i];

	// computing the mean of the data
	dataIndex = 0;		
	for(int dim_i = 0; dim_i < n; dim_i++)
	{	
		double sum = 0.0;
		for(int hyp_i = 0; hyp_i < N*nCarriers; hyp_i++)
		{				
			sum += data[hyp_i * n + dim_i];
			dataIndex += n;
		}
		meanData[dim_i] = (double)sum/(double)(N*nCarriers);
	}

	// subtract mean if solving an inhomogeneous system of equations
	if(!isHomogeneous)
	{
		// centering the data
		for(int dim_i = 0; dim_i < n; dim_i++)
			for(int hyp_i = 0; hyp_i < N*nCarriers; hyp_i++)
				centeredData[hyp_i * n + dim_i] = data[hyp_i * n + dim_i] - meanData[dim_i];
	}
	else // not subtracting the mean otherwise
	{
		for(int dim_i = 0; dim_i < n; dim_i++)
			for(int hyp_i = 0; hyp_i < N*nCarriers; hyp_i++)
				centeredData[hyp_i * n + dim_i] = data[hyp_i * n + dim_i];
	}

	// take svd to get theta from elemental subset
	int N_nCarriers = nCarriers*N;
	dgesvd_("A", "N", &n, &N_nCarriers, centeredData, &n, S, U, &n, V, &N_nCarriers, Work, &Lwork, &Info);

	// extracting theta - last k cols from U
	theta = U + (n-k)*n;

	// compute all projections
	innerProduct(theta,theta_dim,Y,Y_dim,proj_all);

	// compute mean of projections of the hypothesis
	for(int dim_i = 0; dim_i < k; dim_i++)
	{
		alpha[dim_i] = 0.0;
		for(int hyp_i=0; hyp_i < N; hyp_i++)
			for(int const_i = 0; const_i < nCarriers; const_i++)
				alpha[dim_i] += proj_all[((hyp[hyp_i] - 1)*nCarriers + const_i)*k + dim_i];

		alpha[dim_i] /= (double)(N*nCarriers);
	}

	// compute covariance matrices of projections and their inverse, H, invH
	if(isHeteroscedastic)
	{
		for(int pt_i = 0; pt_i < npts*nCarriers; pt_i++)
		{
			int tempMatrix_dim[2] = {n,k};
			matrixMultiply(&(C[pt_i*n*n]), theta, tempMatrix, n, n, k);
			innerProduct(theta,theta_dim,tempMatrix,tempMatrix_dim,H);
			matrixInverse(H,&invH[pt_i*k*k],k,k);
		}
	}

	// compute best carrier vector for every point
	computeBestCarriers(proj_all, proj_best, invH, alpha, dists_sorted, ind_constraints, npts, nCarriers, k, isHeteroscedastic);

	// initialize indices before sorting
	for(int pt_i = 0; pt_i < npts; pt_i++)
		idx_sorted[pt_i] = pt_i;

	// sort points in ascending distances from projected hypothesis
	quickSort(dists_sorted, idx_sorted, 0, npts-1);

	memcpy(inliersS,idx_sorted,ndense*sizeof(int));

	// copy to output
	double* indices_double = (double*)mxMalloc(ndense*sizeof(double));
	for(int i = 0; i < ndense; i++)
		indices_double[i] = (double)inliersS[i] + 1.0;

	mxArray* mxIndices = mxCreateDoubleMatrix(1,ndense,mxREAL);
	plhs[0] = mxIndices;
	memcpy(mxGetPr(mxIndices),indices_double,ndense*sizeof(double));

	mxArray* mxTheta = mxCreateDoubleMatrix(n,k,mxREAL);
	plhs[1] = mxTheta;
	memcpy(mxGetPr(mxTheta),theta,n*k*sizeof(double));

	// deallocate memory
	mxFree(hyp);
	mxFree(data);
	mxFree(meanData);
	mxFree(centeredData);
	mxFree(U);
	mxFree(V);
	mxFree(S);
	mxFree(Work);
	mxFree(proj_best);
	mxFree(proj_all);
	mxFree(inliersS);
	mxFree(indices_double);
	mxFree(alpha);
	mxFree(H);
	mxFree(invH);
	mxFree(dists_sorted);
	mxFree(idx_sorted);
	mxFree(tempMatrix);
	mxFree(ind_constraints);
}
