%% Example script to show usage of gpbM code
% It will be required to build the gpbM source as per the documentation provided with the code
clear;clc;close all
% load data and ground truth
load('line_data.mat','X','gt_X');


% NOTE: In case of multi-carrier problems, the nC carrier vectors from each data point
% should be stacked together. For a multi-carrier problem, the carrier matrix is 
% X = [c_1pt_1, c_2pt_1, ..., c_{nC}pt_1, c_1pt_2, c_2pt_2, ..., c_{nC}pt_2, ...], 
% where c_jpt_i is the j^{th} carrier vector corresponding to the i^{th} data point

%% set up gpbM parameters

epsilon = 0.1;
M = 800; % number of iterations for scale estimation
N = 100; % number of iterations for model estimation
nCarriers = 1; % number of carrier vectors from each data point
isGrassmann = 0; % flag that turns on/off optimization over grassman manifold 
isHeteroscedastic = 0; % flag that turns on/off heteroscedastic processing
isHomogeneous = false; % is the function homogeneous (passes through the origin?)
kernelType = 'Epanechnikov'; % 'Gaussian'; % 
nElem = 2; % number of points to generate an elemental subset

npts = ceil(size(X,2)/nCarriers); % number of data points


degen_func = @isdegenerate_dummy; % degeneracy test


n = size(X,1); % number of dimensions
k = 1; % dimensionality of null space of regression function

% Covariance computation
% For HETEROSCEDASTIC noise, C is an n x n x (npts*nCarriers) array of covariance 
% matrices for each carrier vector (computed using Jacobians)
% For HOMOSCEDASTIC noise, each of the n x n covariance matrices are simply identity
 C = calculateCovariances_homo(X); % function that computes covariances for homoscedastic noise
% NOTE: For multi-carrier problems, the covariance matrices should be stacked along the third 
% dimension, corresponding to X.


%% gpbM 
    
% gpbM step 1: scale estimation
    hypsInScale = generateHypotheses(M, nElem, X,degen_func);
    
    [scale frac bestHypS peakVal inliersScale bestTheta, enclosedPts enclosedFrac] = ...
        estimateScale(X, n, k, hypsInScale, 0, nCarriers,C,nElem,isHomogeneous,epsilon,isHeteroscedastic);

% gpbM step 2: model estimation

% generate hypotheses using only inliers from step 1
    hypsIn = inliersScale(generateHypotheses(N, nElem, X(:,inliersScale),degen_func));
    
    [theta alpha score inliers bestHyp] = estimateModel_mex(X, n, k, hypsIn, C, scale/2, kernelType, ...
        nCarriers,nElem,isHomogeneous,isGrassmann,isHeteroscedastic);
    
% gpbM step 3: inlier/outlier dichotomy
    kde_inliers = dichotomize_mex(X,theta,C,alpha,scale/2,nCarriers,kernelType,isHeteroscedastic);
    fprintf(1,'Detected structure with %d inliers\n',length(kde_inliers));

%% plot inliers corresponding to detected structure
    figure;subplot(1,2,1);plot(X(1,gt_X==1),X(2,gt_X==1),'.','color',[0 0 1]);
    hold on;plot(X(1,gt_X==0),X(2,gt_X==0),'.','color',[0.5,0.5,0.5]);
    title_str = sprintf('Original data: %d inliers, %d outliers',sum(gt_X==1),sum(gt_X==0));
    title(title_str);
    subplot(1,2,2);
    plot(X(1,kde_inliers),X(2,kde_inliers),'r.');
    title_str = sprintf('Detected %d inliers',length(kde_inliers));
    title(title_str);
    axis([-1 3 -1 3]);

